"""
Module to provide the Layer - Stack - Sample classes to build a sample for reflectivity modelling.
The new implementation is based on python dataclasses.

Classes:
ReflFunction - A function class that can be used as a parameter in the other classes.
is_reflfunction - Funtion that checks if an object belongs to the class ReflFunction
ReflBase - Base class for all the physical classes.
LayerBase - Base Layer class.
StackBase - Base Stack class.
SampleBase - Base Sample class.
InstrumentBase - Base Instrument class.

A model should derive its own data classes as base classe of the above like:
::
    @dataclass
    class Instrument(ReflBase):
        wavelength: float = 0.0

Choices should use appropriate Enum to simplify comparison. An example would be:
::
    class Polarization(str, enum.Enum):
        up_up = 'uu'
        down_down = 'dd'
        up_down = 'ud'
        down_up = 'du'

    @dataclass
    class Instrument(ReflBase):
        pol: Polarization = Polarization.up_up
"""

from copy import copy, deepcopy
from dataclasses import dataclass, field, fields
from typing import List

import numpy as np

from .base import ModelParamBase


class ReflBase(ModelParamBase):
    """
    ModelParamBased derived class for use in reflectometry model.
    This class is required for the GUI to recognize classes for use with the Reflectometry plugin.
    """

    Groups = []

    @property
    def _parameters(self):
        # for legacy parameter lookup
        return dict([(fi.name, getattr(self, fi.name)) for fi in fields(self)])

    def _parameter_info(self):
        # field information for all parameters
        return dict([(fi.name, fi) for fi in fields(self)])

    def copy(self, deep=False):
        if deep:
            return deepcopy(self)
        else:
            return copy(self)


@dataclass
class StackBase(ReflBase):
    Layers: List[ReflBase] = field(default_factory=list)
    Repetitions: int = 1

    def resolveLayerParameter(self, name):
        return [getattr(li, name) for li in self.Layers] * self.Repetitions


@dataclass
class SampleBase(ReflBase):
    Stacks: List[StackBase] = field(default_factory=list)
    Ambient: ReflBase = None
    Substrate: ReflBase = None

    # following attribute can be used in sub-class to specify dataclass to return on resolveLayerParameters
    _layer_parameter_class = None
    _is_inverted = False

    def _resolve_parameter(self, obj, key):
        return getattr(obj, key)

    def resolveLayerParameters(self):
        par = {}
        for fi in fields(self.Substrate):
            par[fi.name] = []
            par[fi.name] = [self._resolve_parameter(self.Substrate, fi.name)]
            for stack in self.Stacks:
                par[fi.name] += stack.resolveLayerParameter(fi.name)
            par[fi.name].append(self._resolve_parameter(self.Ambient, fi.name))
        if self._is_inverted:
            for key, values in par.items():
                par[key] = list(reversed(values))
                if key.startswith("sigma"):
                    par[key] = par[key][1:] + [0.0]
        if self._layer_parameter_class:
            return self._layer_parameter_class(**par)
        else:
            return par

    @classmethod
    def _add_sim_method(cls, name, func):
        def method(self, *args):
            nargs = args[:-1] + (self,) + (args[-1],)
            return func(*nargs)

        method.__name__ = "Sim" + name
        setattr(cls, method.__name__, method)

    @classmethod
    def setSimulationFunctions(cls, sim_functions):
        [cls._add_sim_method(k, func) for k, func in sim_functions.items()]

    def __neg__(self):
        """
        The negative of the sample is a representation with inverted layer
        order. This is useful for measurements from the substrate side.
        It takes care of swapping the roughness indices to stay between the same adjecent layers.
        """
        # use a shallow copy as only the bool flag is inverted
        output = self.copy()
        output._is_inverted = not self._is_inverted
        return output


# TODO: Leftovers from old style models, needs refactoring
class ReflFunction:
    def __init__(self, function, validation_args=None, validation_kwargs=None, id=None):
        """Creates the Refl Function given function. The arguments validation_args and
        validation_kwargs will be used to the validate the returned type of the function by passing
        them to function. The variable id should be a unique string to identify the type ReflFunction.
        """
        if validation_args is None and validation_kwargs is None:
            if hasattr(function, "__func__"):
                return
            else:
                raise ValueError("ReflFunction must have validation_args and validation_kwargs arguments")
        self.__func__ = function
        self.validation_args = validation_args
        self.validation_kwargs = validation_kwargs
        self.id = id

    def __new__(cls, *args, **kwargs):
        if len(args) == 1 and isinstance(args[0], ReflFunction):
            return args[0]
        else:
            return super().__new__(cls)

    def __call__(self, *args, **kwargs):
        return self.__func__(*args, **kwargs)

    def validate(self):
        """Function to test that the function returns the anticipated type"""
        return self.__call__(*self.validation_args, **self.validation_kwargs)

    def _check_obj(self, other):
        """Checks the object other so that it fulfills the demands for arithmetic operations."""
        supported_types = [int, float, int, complex, np.float64, np.float32]
        if is_reflfunction(other):
            if self.id != other.id:
                raise TypeError("Two ReflFunction objects must have identical id's to conduct arithmetic operations")
        elif not type(other) in supported_types:
            raise TypeError(
                "%s is not supported for arithmetic operations " % (repr(type(other)))
                + "of a ReflFunction. It has to int, float, long or complex"
            )

    def __mul__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):

            def new_func(*args, **kwargs):
                return self(*args, **kwargs) * other(*args, **kwargs)

        else:

            def new_func(*args, **kwargs):
                return self(*args, **kwargs) * other

        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __rmul__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):

            def new_func(*args, **kwargs):
                return other(*args, **kwargs) * self(*args, **kwargs)

        else:

            def new_func(*args, **kwargs):
                return other * self(*args, **kwargs)

        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __add__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):

            def new_func(*args, **kwargs):
                return self(*args, **kwargs) + other(*args, **kwargs)

        else:

            def new_func(*args, **kwargs):
                return self(*args, **kwargs) + other

        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __radd__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):

            def new_func(*args, **kwargs):
                return other(*args, **kwargs) + self(*args, **kwargs)

        else:

            def new_func(*args, **kwargs):
                return other + self(*args, **kwargs)

        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __sub__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):

            def new_func(*args, **kwargs):
                return self(*args, **kwargs) - other(*args, **kwargs)

        else:

            def new_func(*args, **kwargs):
                return self(*args, **kwargs) - other

        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __rsub__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):

            def new_func(*args, **kwargs):
                return other(*args, **kwargs) - self(*args, **kwargs)

        else:

            def new_func(*args, **kwargs):
                return other - self(*args, **kwargs)

        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __div__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):

            def new_func(*args, **kwargs):
                return self(*args, **kwargs) / other(*args, **kwargs)

        else:

            def new_func(*args, **kwargs):
                return self(*args, **kwargs) / other

        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __rdiv__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):

            def new_func(*args, **kwargs):
                return other(*args, **kwargs) / self(*args, **kwargs)

        else:

            def new_func(*args, **kwargs):
                return other / self(*args, **kwargs)

        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __neg__(self):
        def new_func(*args, **kwargs):
            return -self(*args, **kwargs)

        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __pos__(self):
        def new_func(*args, **kwargs):
            return self(*args, **kwargs)

        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __pow__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):

            def new_func(*args, **kwargs):
                return self(*args, **kwargs) ** other(*args, **kwargs)

        else:

            def new_func(*args, **kwargs):
                return self(*args, **kwargs) ** other

        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __rpow__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):

            def new_func(*args, **kwargs):
                return other(*args, **kwargs) ** self(*args, **kwargs)

        else:

            def new_func(*args, **kwargs):
                return other ** self(*args, **kwargs)

        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)


def is_reflfunction(obj):
    """Convenience function to determine whether obj belongs to the ReflFunction class.
    Return boolean.
    """
    return obj.__class__.__name__ == "ReflFunction"


def cast_to_array(list_of_obj, *args, **kwargs):
    """Casts a list_of_obj, can be a number or an ReflFunction, into a list of evaluated values"""
    id = ""
    shape = False
    ret_list = []
    for obj in list_of_obj:
        if is_reflfunction(obj):
            if id == "":
                id = obj.id
                ret_list.append(obj(*args, **kwargs))
                # Check if we have got an array
                if not np.isscalar(ret_list[-1]):
                    shape = ret_list[-1].shape
            elif id == obj.id:
                ret_list.append(obj(*args, **kwargs))
            else:
                TypeError("Two ReflFunction objects must have identical id's in order to merge them into an array")
        else:
            # We assume that this is an object that can be transformed into an array later on.
            ret_list.append(obj)
    # if we have an array make sure that all the objects have the same shape
    if shape:
        base_array = np.ones(shape)
        nret_list = []
        for item in ret_list:
            if np.isscalar(item):
                nret_list.append(item * base_array)
            else:
                nret_list.append(item)
        ret_list = nret_list

    return np.array(ret_list)


def harm_sizes(ar, shape, dtype=np.float64):
    """Utility function to add an additional axis if needed to fulfill the size in shape"""
    ar = np.array(ar, dtype=dtype)
    if shape is None:
        return ar
    elif len(ar.shape) < len(shape):
        return np.array(ar[..., np.newaxis] * np.ones(shape), dtype=dtype)
    elif ar.shape == shape:
        return ar
    else:
        raise TypeError("The size of the array, %s, can not be changed to shape %s" % (ar.shape, shape))



"""
Define base classes that can be used in models for prameterization. This classes are model independent.
"""

import ast
import dataclasses
import inspect
import logging
import sys

from dataclasses import dataclass, field, fields
from enum import Enum

# change of signature introduced in python 3.10.1
if sys.version_info >= (3, 10, 1):
    _field_init_real = dataclasses._field_init

    def _field_init(f, frozen, locals, self_name):
        return _field_init_real(f, frozen, locals, self_name, False)

else:
    _field_init = dataclasses._field_init


def _custom_init_fn(fieldsarg, frozen, has_post_init, self_name, globals):
    """
    _init_fn from dataclasses adapted to accept additional keywords.
    See dataclasses source for comments on code.
    """
    seen_default = False
    for f in fieldsarg:
        if f.init:
            if not (f.default is dataclasses.MISSING and f.default_factory is dataclasses.MISSING):
                seen_default = True
            elif seen_default:
                raise TypeError(f"non-default argument {f.name!r} " "follows default argument")

    if sys.version_info >= (3, 12, 0):
        # changed behavior in dataclasses
        locals = {f"__dataclass_type_{f.name}__": f.type for f in fieldsarg}
        locals.update(
            {
                "__dataclass_HAS_DEFAULT_FACTORY__": dataclasses._HAS_DEFAULT_FACTORY,
                "__dataclass_builtins_object__": object,
            }
        )
    else:
        locals = {f"_type_{f.name}": f.type for f in fieldsarg}
        locals.update({"MISSING": dataclasses.MISSING, "_HAS_DEFAULT_FACTORY": dataclasses._HAS_DEFAULT_FACTORY})

    body_lines = []
    for f in fieldsarg:
        line = _field_init(f, frozen, locals, self_name)
        if line:
            body_lines.append(line)

    if has_post_init:
        params_str = ",".join(f.name for f in fieldsarg if f._field_type is dataclasses._FIELD_INITVAR)
        body_lines.append(f"{self_name}.{dataclasses._POST_INIT_NAME}({params_str})")

    # processing of additional user keyword arguments
    body_lines += ["setattr(self, '_unused_kwds', user_kwds)"]

    arg_list_genx = [self_name, ] + [dataclasses._init_param(f) for f in fieldsarg if f.init] + ["**user_kwds"]
    if sys.version_info >= (3, 13, 0):
        # Make sure all body_lines are indented.
        body_lines = [f'  {l.strip()}' for l in body_lines]
        func_builder = dataclasses._FuncBuilder(globals)
        func_builder.add_fn("__init__", arg_list_genx, body_lines, locals=locals, return_type=None)
        func_builder.unconditional_adds['__init__']=True
        return func_builder
    else:
        return dataclasses._create_fn(
            "__init__",
            arg_list_genx,
            body_lines,
            locals=locals,
            globals=globals,
            return_type=None,
        )


class ModelParamMeta(type):
    """
    Metaclass that overwrites a classes __init__ method to make sure dataclass is generated without
    own __init__ method and thus does ignore additional keyword arguments.
    """

    def __new__(meta, name, bases, attrs):
        cls = super().__new__(meta, name, bases, attrs)

        if (
            "__annotations__" in attrs
            and len([k for k in attrs["__annotations__"].keys() if not k.startswith("_")]) > 0
        ):
            as_dataclass = dataclass(cls)
            fieldsarg = getattr(as_dataclass, dataclasses._FIELDS)

            # Generate custom __init__ method that allows arbitrary extra keyword arguments
            has_post_init = hasattr(as_dataclass, dataclasses._POST_INIT_NAME)
            # Include InitVars and regular fields (so, not ClassVars).
            flds = [f for f in fieldsarg.values() if f._field_type in (dataclasses._FIELD, dataclasses._FIELD_INITVAR)]
            init_fun = _custom_init_fn(flds, False, has_post_init, "self", globals())
            if sys.version_info >= (3, 13, 0):
                init_fun.add_fns_to_class(cls)
            else:
                setattr(cls, "__init__", init_fun)
        # adding documentation for class into module doc-string
        # this leads to much cleaner code in model module as beginning of file only contains
        # a general introduction and each dataclass gets documented where it's defined.
        module = inspect.getmodule(cls)
        if module is not None and cls.__doc__ is not None:
            # create suitable doc string automatically, but only for non-interactively created classes
            doc = inspect.cleandoc(cls.__doc__)
            docout = f"{cls.__name__}\n"
            docout += "~" * len(cls.__name__) + "\n"
            sig = eval("inspect.signature(cls)", module.__dict__, {"cls": cls, "inspect": inspect})
            units = getattr(cls, "Units", {})
            call_params = []
            for ni, si in sig.parameters.items():
                if ni == "user_kwds":
                    continue
                if ni in units and units[ni].strip() != "":
                    call_params.append(f"{ni}={si.default!r} ({units[ni]})")
                else:
                    call_params.append(f"{ni}={si.default!r}")
            call_string = f'{cls.__name__}({", ".join(call_params)})'
            docout += f"``{call_string}``\n\n"
            docout += doc
            module.__doc__ += docout + "\n\n"

        return cls


class ModelParamBase(metaclass=ModelParamMeta):
    """
    Base class for all reflectometry dataclass objects. It uses the ReflMeta class to
    allow arbitrary ununsed keyword arguments (for switching models) and stores
    the call strings used when creating a class instance to allow re-creation
    within the GUI.

    In addition to the dataclass attributes a subclass can define a Units
    dictionary that can be used in GUI and documentation to annotate the
    physical unit used for the parameter.
    """

    Units = {}
    CREATES_OWN_SETTER = 'CREATES_OWN_SETTER' # indicates the setter function should not be created for child class
    DEFAULT_FIT_PARAMS = [] # Classes can define default attributes that are filled in the grid

    def __post_init__(self):
        self._generate_setters()
        self._generate_getters()
        # record the line number in the script when executed by exec, see model.Model.compile_script
        self._orig_params = {}
        self._ca = {}
        for fi in fields(self):
            if fi.name not in self.Units:
                self.Units[fi.name] = ""
            self._orig_params[fi.name] = getattr(self, fi.name)
            if inspect.isclass(fi.type) and not isinstance(getattr(self, fi.name), fi.type):
                # convert parameter to correct type
                # TODO: implement for more complex types like Union[a,b], check above works for both
                setattr(self, fi.name, fi.type(getattr(self, fi.name)))
        direct_call_context = inspect.stack()[2]
        if direct_call_context.filename.endswith(".py"):
            # the class was generated in some other scope then the script, don't record line number
            self._lno_context = None
            return
        else:
            self._lno_context = inspect.stack()[2].lineno - 1

    def _extract_callpars(self, source):
        if self._lno_context is None:
            return
        # try to extract call arguments for the keywords
        context = [source.splitlines()[self._lno_context]]
        call_lines = "".join(context)
        ca = {}
        try:
            p = ast.parse(call_lines)
            kws = p.body[0].value.keywords
            for kw in kws:
                if hasattr(ast, "get_source_segment"):
                    ca[kw.arg] = ast.get_source_segment(call_lines, kw.value)
                else:
                    ca[kw.arg] = call_lines[kw.value.col_offset : kw.value.end_col_offset]
        except Exception:
            pass
        for fi in fields(self):
            if not fi.name in ca:
                ca[fi.name] = repr(self._orig_params[fi.name])
        self._ca = ca

    @staticmethod
    def _get_setter(object, par):
        def set_func(value):
            setattr(object, par, value)

        set_func.__name__ = "set" + par.capitalize()
        return set_func

    @staticmethod
    def _get_real_setter(object, par):
        def set_func(value):
            setattr(object, par, value + getattr(object, par).imag * 1j)

        set_func.__name__ = "set" + par.capitalize() + "real"
        return set_func

    @staticmethod
    def _get_imag_setter(object, par):
        def set_func(value):
            setattr(object, par, value * 1j + getattr(object, par).real)

        set_func.__name__ = "set" + par.capitalize() + "imag"
        return set_func

    def _generate_setters(self):
        for fi in fields(self):
            if fi.type is self.CREATES_OWN_SETTER:
                continue
            par = str(fi.name)

            set_func = ModelParamBase._get_setter(self, par)
            setattr(self, set_func.__name__, set_func)

            if fi.type is complex or complex in getattr(fi.type, "__args__", ()):
                set_real_func = ModelParamBase._get_real_setter(self, par)
                setattr(self, set_real_func.__name__, set_real_func)

                set_imag_func = ModelParamBase._get_imag_setter(self, par)
                setattr(self, set_imag_func.__name__, set_imag_func)

    @staticmethod
    def _get_getter(object, par):
        def get_func():
            return getattr(object, par)

        get_func.__name__ = "get" + par.capitalize()
        return get_func

    @staticmethod
    def _get_real_getter(object, par):
        def get_func():
            return getattr(object, par).real

        get_func.__name__ = "get" + par.capitalize() + "real"
        return get_func

    @staticmethod
    def _get_imag_getter(object, par):
        def get_func():
            return getattr(object, par).imag

        get_func.__name__ = "get" + par.capitalize() + "imag"
        return get_func

    def _generate_getters(self):
        for fi in fields(self):
            if fi.type is self.CREATES_OWN_SETTER:
                continue
            par = str(fi.name)
            get_func = ModelParamBase._get_getter(self, par)
            setattr(self, get_func.__name__, get_func)

            if fi.type is complex:
                get_real_func = ModelParamBase._get_real_getter(self, par)
                setattr(self, get_real_func.__name__, get_real_func)

                get_imag_func = ModelParamBase._get_imag_getter(self, par)
                setattr(self, get_imag_func.__name__, get_imag_func)

    def _repr_call(self):
        output = self.__class__.__name__ + "("
        for fi in fields(self):
            call = self._ca.get(fi.name, None)
            if call is not None:
                output += f"{fi.name}={call}, "
            else:
                output += f"{fi.name}={getattr(self, fi.name)!r}, "
        return output[:-2] + ")"


class AltStrEnum(str, Enum):
    """
    A string based Enum that allows to create equivalent values from different strings.
    Therefore, the user can select various values but the code compares just to one item.

    Any alternative attributes have to start with "alternate??_" and end with `name`, where
    `name` is the original attribute to be doublicated.
    """

    @classmethod
    def _missing_(cls, value):
        # allow integer indices for backwards compatibility of some models
        # order or items is according to order of definition in class
        values = [e.value for e in cls]
        if type(value) is int and value >= 0 and value < len(values):
            return cls(values[value])
        else:
            return super()._missing_(value)

    def __init__(self, string):
        if self.name.startswith("alternate"):
            self.alt_for = self.name.split("_", 1)[1]
        else:
            self.alt_for = None

    def __eq__(self, other):
        try:
            # if an item can't be converted to this Enum, it can't be equal
            other = self.__class__(other)
        except ValueError:
            return False
        if getattr(other, "alt_for", None) is not None:
            return self.name == other.alt_for or str.__eq__(self, other)
        elif getattr(self, "alt_for", None) is not None:
            return self.alt_for == other.name or str.__eq__(self, other)
        else:
            return str.__eq__(self, other)

    def __str__(self):
        return self.value

    def __repr__(self):
        return f'"{self.value}"'

    def __hash__(self):
        return str.__hash__(self.value)



"""
=================
FrequencyAnalysis
=================

A plugin to show reflectivity frequency components by using FFT/CWT of reflectivity corrected by critical q-values.

Written by Artur Glavic
"""

import wx

from numpy import isnan, linspace

from genx.gui.plotpanel import BasePlotConfig, PlotPanel

from .. import add_on_framework as framework
from .help_modules.frequency_analysis import TransformType, transform


class FAPlotConfig(BasePlotConfig):
    section = "frequency analysis plot"


class FAPlotPanel(wx.Panel):
    """Widget for plotting the frequency analysis of datasets."""

    def __init__(self, parent, plugin, id=-1, color=None, dpi=None, style=wx.NO_FULL_REPAINT_ON_RESIZE, **kwargs):
        wx.Panel.__init__(self, parent)
        self.plot = PlotPanel(self, -1, color, dpi, FAPlotConfig, style, **kwargs)
        self.plugin = plugin

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.plot, 1, wx.EXPAND | wx.GROW | wx.ALL)
        hsizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(hsizer, 0, wx.FIXED_MINSIZE)
        lsizer = wx.BoxSizer(wx.VERTICAL)
        hsizer.Add(lsizer, 1, wx.EXPAND | wx.GROW)
        csizer = wx.BoxSizer(wx.VERTICAL)
        hsizer.Add(csizer, 1, wx.EXPAND | wx.GROW)
        rsizer = wx.BoxSizer(wx.VERTICAL)
        hsizer.Add(rsizer, 1, wx.EXPAND | wx.GROW)
        rlabels = wx.BoxSizer(wx.VERTICAL)
        hsizer.Add(rlabels, 1, wx.EXPAND | wx.GROW)

        self.plot.update(None)
        self.plot.ax = self.plot.figure.add_subplot(111)
        box = self.plot.ax.get_position()
        self.plot.ax.set_position([box.x0, box.y0, box.width * 0.95, box.height])
        self.plot.ax.set_autoscale_on(True)
        self.plot.update = self.Plot
        self.SetSizer(sizer)
        self.plot_dict = {}

        self.transform_log = wx.CheckBox(self, -1, "log(R)")
        lsizer.Add(self.transform_log, 0, wx.FIXED_MINSIZE)
        self.transform_Q4 = wx.CheckBox(self, -1, "R Q^4")
        self.transform_Q4.SetValue(True)
        lsizer.Add(self.transform_Q4, 0, wx.FIXED_MINSIZE)
        self.use_derivative = wx.CheckBox(self, -1, "dR/dQ")
        lsizer.Add(self.use_derivative, 0, wx.FIXED_MINSIZE)

        self.tt = {}
        for tt in list(TransformType):
            self.tt[tt] = wx.RadioButton(self, -1, label=tt.name)
            csizer.Add(self.tt[tt], 0, wx.FIXED_MINSIZE)
            self.tt[tt].Bind(wx.EVT_RADIOBUTTON, self.Plot)
        self.tt[TransformType.fourier_transform].SetValue(True)

        self.Qc = wx.SpinCtrlDouble(self, -1, value="0.05", min=0.00, max=1.5, inc=0.001)
        self.Qc.SetDigits(5)
        rlabels.Add(wx.StaticText(self, -1, "Qc"), 0, wx.FIXED_MINSIZE)
        rsizer.Add(self.Qc, 0, wx.FIXED_MINSIZE)

        self.WLscale = wx.SpinCtrlDouble(self, -1, value="0.5", min=0.00, max=1.0, inc=0.01)
        self.WLscale.SetDigits(3)
        rlabels.Add(wx.StaticText(self, -1, "WL-scale"), 0, wx.FIXED_MINSIZE)
        rsizer.Add(self.WLscale, 0, wx.FIXED_MINSIZE)

        self.transform_log.Bind(wx.EVT_CHECKBOX, self.Plot)
        self.transform_Q4.Bind(wx.EVT_CHECKBOX, self.Plot)
        self.use_derivative.Bind(wx.EVT_CHECKBOX, self.Plot)
        self.Qc.Bind(wx.EVT_SPINCTRLDOUBLE, self.Plot)
        self.Qc.Bind(wx.EVT_TEXT_ENTER, self.Plot)
        self.WLscale.Bind(wx.EVT_SPINCTRLDOUBLE, self.Plot)
        self.WLscale.Bind(wx.EVT_TEXT_ENTER, self.Plot)

    def SetZoom(self, active=False):
        return self.plot.SetZoom(active)

    def GetZoom(self):
        return self.plot.GetZoom()

    def Plot(self, event=None):
        """Plot(self) --> None

        Plotting the sample Sample.
        """
        colors = ["b", "r", "g", "c", "m", "y", "k"]
        # self.plot_dict = model.sample.SimSLD(None, model.inst)
        while len(self.plot.ax.lines) > 0:
            self.plot.ax.lines[0].remove()
        i = 0
        data = self.plugin.GetModel().get_data()

        d = linspace(10.0, 2500.0, 250)
        # New style sim function with one sld for each simulation
        options = dict(
            Qc=float(self.Qc.GetValue()),
            logI=self.transform_log.IsChecked(),
            Q4=self.transform_Q4.IsChecked(),
            derivate=self.use_derivative.IsChecked(),
            wavelet_scaling=self.WLscale.GetValue(),
        )
        for tt, ctrl in self.tt.items():
            if ctrl.GetValue():
                options["trans_type"] = tt
                break
        for i, di in enumerate(data):
            if i % 2 != 0:
                continue
            if di.show:
                try:
                    if not isnan(di.y).all() and (di.y > 0).any():
                        _, mag = transform(di.x, di.y, derivN=3, Qmin=0.08, Qmax=None, D=d, **options)
                        self.plot.ax.plot(
                            d,
                            mag,
                            color=di.data_color,
                            ls=di.data_linetype,
                            lw=di.data_linethickness,
                            marker=di.data_symbol,
                            ms=di.data_symbolsize,
                            label="data " + di.name,
                        )

                    if di.y_sim is not None and not isnan(di.y_sim).all() and (di.y_sim > 0).any():
                        _, mag = transform(di.x, di.y_sim, derivN=3, Qmin=0.08, Qmax=None, D=d, **options)
                        self.plot.ax.plot(
                            d,
                            mag,
                            color=di.sim_color,
                            ls=di.sim_linetype,
                            lw=di.sim_linethickness,
                            marker=di.sim_symbol,
                            ms=di.sim_symbolsize,
                            label="sim " + di.name,
                        )
                except KeyError:
                    pass

        self.plot.ax.legend(loc="upper right", framealpha=0.5, fontsize="small", ncol=1)  # bbox_to_anchor=(1, 0.5),

        self.plot.ax.yaxis.label.set_text("magnitude [a.u.]")
        self.plot.ax.xaxis.label.set_text("d [Å]")
        wx.CallAfter(self.plot.flush_plot)
        self.plot.AutoScale()

    def ReadConfig(self):
        """ReadConfig(self) --> None

        Reads in the config file
        """
        return self.plot.ReadConfig()

    def GetYScale(self):
        """GetYScale(self) --> String

        Returns the current y-scale in use. Currently the string
        'log' or 'linear'. If the axes does not exist it returns None.
        """
        if self.plot.ax:
            return self.plot.ax.get_yscale()
        else:
            return None

    def GetXScale(self):
        """GetXScale(self) --> String

        Returns the current x-scale in use. Currently the string
        'log' or 'linear'. If the axes does not exist it returns None.
        """
        if self.plot.ax:
            return self.plot.ax.get_xscale()
        else:
            return None


class Plugin(framework.Template):
    def __init__(self, parent):
        framework.Template.__init__(self, parent)
        self.parent = parent

        # Create the SLD plot
        FA_plot_panel = self.NewPlotFolder("Frequency Analysis")
        FA_sizer = wx.BoxSizer(wx.HORIZONTAL)
        FA_plot_panel.SetSizer(FA_sizer)
        self.FA_plot = FAPlotPanel(FA_plot_panel, self)
        FA_sizer.Add(self.FA_plot, 1, wx.EXPAND | wx.GROW | wx.ALL)
        FA_plot_panel.Layout()

    def OnSimulate(self, event):
        """OnSimulate(self, event) --> None

        Updates stuff after simulation
        """
        # Calculate and update the sld plot
        wx.CallAfter(self.FA_plot.Plot)

    def OnFittingUpdate(self, event):
        """OnSimulate(self, event) --> None

        Updates stuff during fitting
        """
        wx.CallAfter(self.FA_plot.Plot())



from genx.version import __version__

mv = int(__version__.split(".")[1])

RELEASE_HEAD = """
Changes since previous release
==============================
"""

ALPHA_NOTE = """
CAUTION
=======
This is an alpha release version. It may provide new functionality but likely contains new bugs
that may break important functionality. Please don't use it in production environments.

Please be sure to report any unexpected behavior and issues as
[ticket on SourceForge](https://sourceforge.net/p/genx/tickets/) or 
[via e-mail](mailto:artur.glavic@psi.ch).
"""

BETA_NOTE = """
NOTE
====
This is a beta release version. It may provide new functionality but can contain new bugs
that could break some functionality.

Please be sure to report any unexpected behavior and issues as
[ticket on SourceForge](https://sourceforge.net/p/genx/tickets/) or 
[via e-mail](mailto:artur.glavic@psi.ch).
"""

with open("release_notes.md", "w") as fh:
    found_first = 0
    if "a" in __version__:
        fh.write(ALPHA_NOTE)
    elif "b" in __version__:
        fh.write(BETA_NOTE)
    fh.write(RELEASE_HEAD)
    for l in open("README.txt", "r").readlines():
        if f"Changes 3." in l:
            if found_first:
                break
            else:
                found_first = 3
        if found_first == 1:
            fh.write(l)
        elif found_first:
            found_first -= 1



"""
=============
LayerGraphics
=============

A plugin to create sketch graphics to vizualize the layer structure.
The result is an SVG image with automatically generated colors and sizes.

Thicknesses can either be linear or mapped through a scaling funciton to
avoid large variation if thicknesses are too different.
"""

from dataclasses import dataclass

import svgwrite
import wx

from numpy import array, log, pi, sqrt

from genx.models.lib.refl_base import SampleBase, StackBase

from .. import add_on_framework as framework

try:
    # noinspection PyUnresolvedReferences
    from typing import Dict, List
except ImportError:
    List = object


@dataclass
class LColor:
    r: float
    g: float
    b: float

    def __add__(self, other):
        out = LColor(self.r + other.r, self.g + other.g, self.b + other.b)

    def normalize(self):
        norm = sqrt((array([self.r, self.g, self.b]) ** 2).sum())
        self.r /= norm
        self.g /= norm
        self.b /= norm

    @property
    def side(self):
        # return lighter color for side of 3d block
        return LColor(min(1.0, self.r + 0.35), min(1.0, self.g + 0.35), min(1.0, self.b + 0.35))

    @property
    def top(self):
        # return lighter color for side of 3d block
        return LColor(min(1.0, self.r + 0.2), min(1.0, self.g + 0.2), min(1.0, self.b + 0.2))

    def __str__(self):
        return f"rgb({self.r*255:.0f}, {self.g*255:.0f}, {self.b*255:.0f})"

    def get_wx(self):
        return wx.Colour(int(self.r * 255), int(self.g * 255), int(self.b * 255))


@dataclass
class LInfo:
    thickness: float
    scale: float
    color: LColor
    name: str
    sld: str

COLORS = [
    LColor(1.0, 0.0, 0.0),
    LColor(0.0, 1.0, 0.0),
    LColor(0.0, 0.0, 1.0),
    LColor(1.0, 1.0, 0.0),
    LColor(1.0, 0.0, 1.0),
    LColor(0.0, 1.0, 1.0),
]


class BlockGenerator:
    stacks: List[StackBase]
    rescale: bool  # apply the mapping function to reduce visible variation of size
    is_xray: bool # use x-ray as basis for sld calculation
    show_all: bool  # show all layers in a repeating stack separately
    show_one: bool  # show just one sequence in a repreating stack when above 1 repetition

    dmin: float
    dmax: float
    dtotal: float

    repetitions: List[int]
    thicknesses: List[List[float]]
    scaled: List[List[float]]

    min_rescale = 30.0
    max_rescale = 100.0

    def __init__(self, stacks: List[StackBase], layer_names: Dict[int, str],
                 is_xray=True, rescale=True, show_all=False, show_one=False):
        self.stacks = [si for si in stacks if len(si.Layers) > 0]
        self.layer_names = layer_names
        self.is_xray = is_xray
        self.rescale = rescale
        self.show_all = show_all
        self.show_one = show_one
        self.get_sample_range()

    def rescale_thickness(self, di):
        ldi = log(di)
        ldmin = log(self.dmin)
        ldmax = log(self.dmax)
        if ldmin == ldmax:
            # don't resace if all thicknesses are the same
            return di
        ld = (ldi - ldmin) / (ldmax - ldmin)
        sd = self.min_rescale + (self.max_rescale - self.min_rescale) * ld
        return sd

    def get_sample_range(self):
        # return total thickness, smallest and largest layer
        dmin = 1.0e5
        dmax = 0.0
        dtotal = 0.0

        repetitions = []
        thicknesses = []
        color_slds = []
        names = []
        slds = []

        for stack in self.stacks:
            repetitions.append(int(stack.Repetitions))
            thicknesses.append([li.d for li in stack.Layers])
            names.append([self.layer_names.get(id(li), '').replace('_', ' ') for li in stack.Layers])
            if self.is_xray:
                slds.append([f'{li.dens*li.f.real:.2f} rₑ/Å³' for li in stack.Layers])
                color_slds.append([abs(li.dens*li.f) for li in stack.Layers])
            else:
                slds.append([f'{li.dens*li.b.real*10.:.2f} 10⁻⁶ Å⁻²' for li in stack.Layers])
                color_slds.append([abs(li.dens * li.b) for li in stack.Layers])
            if len(thicknesses[-1]) > 0:
                dmin = min(dmin, min(thicknesses[-1]))
                dmax = max(dmax, max(thicknesses[-1]))

        unique_ids = {}
        color_ids = []
        cid = 0
        for group_slds in color_slds:
            group_ids = []
            for csld in group_slds:
                csld = round(csld, 4)
                if csld not in unique_ids:
                    unique_ids[csld] = cid
                    cid += 1
                group_ids.append(unique_ids[csld])
            color_ids.append(group_ids)

        self.dmin = dmin
        self.dmax = dmax
        self.repetitions = repetitions
        self.color_ids = color_ids

        if self.rescale:
            sthicknesses = [list(map(self.rescale_thickness, ti)) for ti in thicknesses]
        else:
            sthicknesses = thicknesses
        self.thicknesses = thicknesses
        self.names = names
        self.slds = slds
        self.scaled = sthicknesses

        for ri, ti in zip(repetitions, sthicknesses):
            dsequence = sum(ti)
            if ri < 2:
                dtotal += dsequence
            elif self.show_one:
                dtotal += dsequence + dsequence / 2
            elif self.show_all or ri == 2:
                dtotal += ri * dsequence
            else:
                dtotal += 2 * dsequence + dsequence / 2
        self.dtotal = dtotal

    def get_blocks(self):
        """
        Returns a list of blocks with either a thickness or list of thicknesses
        """
        output = []
        for ri, ti, si, cidi, ni, di in zip(self.repetitions, self.thicknesses, self.scaled,
                                            self.color_ids, self.names, self.slds):
            infos = [LInfo(ti_j, si_j, COLORS[cid_j % len(COLORS)], ni_j, di_j)
                     for ti_j, si_j, cid_j, ni_j, di_j in zip(ti, si, cidi, ni, di)]
            dsequence = sum(si)
            if ri < 2:
                output += infos
            elif self.show_one:
                output.append(infos + [LInfo(-ri, dsequence / 2, LColor(1.0, 1.0, 1.0), '', '')])
            elif self.show_all or ri == 2:
                output.append(ri * infos)
            else:
                output.append(infos + [LInfo(-ri, dsequence / 2, LColor(1.0, 1.0, 1.0), '', '')] + infos)
        return output

    def __repr__(self):
        blocks = self.get_blocks()
        bstr = "substrate |"
        sstr = " |"
        for bi in blocks:
            if isinstance(bi, list):
                bstr += f"|"
                sstr += f"|"
                for bij in bi:
                    sstr += f" {bij.scale:.0f} |"
                    if bij.thickness < 0:
                        bstr = bstr[:-1]
                        bstr += " (x%i) " % (-bij.thickness)
                    else:
                        bstr += f" {bij.thickness:.0f} |"
                bstr += f"|"
                sstr += f"|"
            else:
                bstr += f" {bi.thickness:.0f} |"
                sstr += f" {bi.scale:.0f} |"
        bstr += " ambient"
        return f"BlockGenerator(\n               {bstr}\n               {sstr})"


class SVGenerator:
    block_generator: BlockGenerator
    svg: svgwrite.Drawing
    use3d: bool

    unit_precision: int  # number of digits behind decimal point for thickness label
    unit_nm: bool  # use nm instead of angstrom as unit

    view_box = array([100, 200])
    move_3d = 8.0
    vanishing_point = (175.0, -15.0)
    fontsize = 10

    def __init__(
        self,
        sample: SampleBase,
        layer_names,
        rescale=True,
        show_all=False,
        show_one=False,
        show_names=False,
        show_slds=False,
        use3d=True,
        unit_precision=1,
        unit_nm=True,
        fontsize=10,
        is_xray=True,
    ):
        self.use3d = use3d
        self.fontsize = fontsize
        self.unit_precision = unit_precision
        self.unit_nm = unit_nm
        self.show_names = show_names
        self.show_slds = show_slds
        self.block_generator = BlockGenerator(sample.Stacks, layer_names=layer_names, is_xray=is_xray,
                                              rescale=rescale, show_all=show_all, show_one=show_one)
        self.create_svg()

    def to_vp(self, point0):
        """
        Calculate the x,y-values for a shift from a starting point (x0,y0)
        towards the vanishing point by an x-offset of self.move_3d.
        """
        px, py = self.view_box * 0.01
        x0, y0 = point0
        dx = self.move_3d
        vx, vy = self.vanishing_point
        relx = dx / (vx - x0)  # releative movement towrads vanishing point
        dy = (vy - y0) * relx
        return ((x0 + dx) * px, (y0 + dy) * py)

    def tlabel(self, thickness):
        if self.unit_nm:
            output = f"{thickness/10:.{self.unit_precision}f} nm"
        else:
            output = f"{thickness:.{self.unit_precision}f} Å"
        return output

    def add_rect(self, pos: float, si: float, color: LColor = LColor(0.0, 0.0, 0.0), in_stack=False, width=90):
        px, py = self.view_box * 0.01
        if in_stack:
            x = 5 + self.fontsize
            w = width - self.fontsize
        else:
            x = 5
            w = width
        if self.use3d:
            w -= 10
            points = [
                ((x + w) * px, pos * py),
                ((x + w) * px, (pos + si) * py),
                self.to_vp(((x + w), (pos + si))),
                self.to_vp(((x + w), pos)),
            ]
            side = self.svg.polygon(points=points, stroke="black", stroke_width=0.5, fill=str(color.side))
            self.svg.add(side)
        rect = self.svg.rect((x * px, pos * py), (w * px, si * py), stroke="black", stroke_width=0.5, fill=str(color))
        self.svg.add(rect)

    def create_svg(self):
        px, py = self.view_box * 0.01
        bg = self.block_generator
        self.svg = svgwrite.Drawing(
            "genx_model.svg", size=("4cm", "8cm"), viewBox=f"0 0 {self.view_box[0]} {self.view_box[1]}"
        )
        vscale = bg.dtotal / 85.0  # 90% of image height to be used for layers

        blocks = bg.get_blocks()

        if self.use3d:
            tc = 45
        else:
            tc = 50

        pos = 90.0
        self.add_rect(pos, 100, LColor(0, 0, 0))
        bc = LColor(0, 0, 0)
        for bi in blocks:
            if isinstance(bi, list):
                block_length = sum([bij.scale / vscale for bij in bi])
                self.add_rect(pos - block_length, block_length, LColor(0.75, 0.75, 0.75), width=10 + self.fontsize)
                for bij in bi:
                    dij = bij.scale / vscale
                    pos -= dij
                    if bij.thickness < 0:
                        if self.use3d:
                            # add surface polygon
                            shifted = self.to_vp((85, pos + dij))
                            points = [
                                ((5 + self.fontsize) * px, (pos + dij) * py),
                                (85 * px, (pos + dij) * py),
                                shifted,
                                (shifted[0] - (85 - 5 - self.fontsize) * px, shifted[1]),
                            ]
                            side = self.svg.polygon(points=points, stroke="black", stroke_width=0.5, fill=str(bc.top))
                            self.svg.add(side)
                        paragraph = self.svg.add(self.svg.g(font_size=self.fontsize))
                        paragraph.add(
                            self.svg.text(
                                " ... ",
                                (tc * px, (pos + dij / 3.0) * py),
                                text_anchor="middle",
                                dominant_baseline="middle",
                            )
                        )
                        paragraph = self.svg.add(
                            self.svg.g(font_size=self.fontsize, transform=f"translate(5, {(pos+dij)*py}) rotate(-90)")
                        )
                        paragraph.add(
                            self.svg.text(
                                f"(x%i)" % (-bij.thickness), (0, 0), text_anchor="middle", dominant_baseline="hanging"
                            )
                        )
                    else:
                        bc = bij.color
                        self.add_rect(pos, dij, bc, in_stack=True)
                        paragraph = self.svg.add(self.svg.g(font_size=self.fontsize))
                        paragraph.add(
                            self.svg.text(
                                self.tlabel(bij.thickness),
                                (tc * px, (pos + dij / 2.0) * py),
                                text_anchor="middle",
                                dominant_baseline="middle",
                            )
                        )
                        if self.show_names:
                            paragraph.add(
                                self.svg.text(
                                    bij.name,
                                    (tc * px, (pos + dij / 2.0) * py-self.fontsize*1.2),
                                    text_anchor="middle",
                                    dominant_baseline="middle",
                                )
                            )
                        if self.show_slds:
                            paragraph.add(
                                self.svg.text(
                                    bij.sld,
                                    (tc * px, (pos + dij / 2.0) * py+self.fontsize*1.2),
                                    text_anchor="middle",
                                    dominant_baseline="middle",
                                )
                            )
            else:
                di = bi.scale / vscale
                pos -= di
                bc = bi.color
                self.add_rect(pos, di, bc)
                paragraph = self.svg.add(self.svg.g(font_size=self.fontsize))
                paragraph.add(
                    self.svg.text(
                        self.tlabel(bi.thickness),
                        (tc * px, (pos + di / 2.0) * py),
                        text_anchor="middle",
                        dominant_baseline="middle",
                    )
                )
                if self.show_names:
                    paragraph.add(
                        self.svg.text(
                            bi.name,
                            (tc*px, (pos+di/2.0)*py-self.fontsize*1.2),
                            text_anchor="middle",
                            dominant_baseline="middle",
                            )
                        )
                if self.show_slds:
                    paragraph.add(
                        self.svg.text(
                            bi.sld,
                            (tc*px, (pos+di/2.0)*py+self.fontsize*1.2),
                            text_anchor="middle",
                            dominant_baseline="middle",
                            )
                        )
        if self.use3d:
            # add surface polygon
            shifted = self.to_vp((85, pos))
            points = [(5 * px, pos * py), (85 * px, pos * py), shifted, (shifted[0] - 80 * px, shifted[1])]
            side = self.svg.polygon(points=points, stroke="black", stroke_width=0.5, fill=str(bc.top))
            self.svg.add(side)

    def add_rect_gc(
        self,
        gc: wx.GraphicsContext,
        pos: float,
        si: float,
        color: LColor = LColor(0.0, 0.0, 0.0),
        in_stack=False,
        width=90,
    ):
        px, py = self.view_box * 0.01
        gx, gy = gc.GetSize()
        gc_scale = min(gx / self.view_box[0], gy / self.view_box[1])

        if in_stack:
            x = 5 + self.fontsize
            w = width - self.fontsize
        else:
            x = 5
            w = width
        if self.use3d:
            w -= 10
            points = [
                ((x + w) * px, pos * py),
                ((x + w) * px, (pos + si) * py),
                self.to_vp(((x + w), (pos + si))),
                self.to_vp(((x + w), pos)),
            ]
            self.draw_polygon_gc(gc, points, color.side)
        gc.SetPen(wx.BLACK_PEN)
        gc.SetBrush(wx.Brush(color.get_wx(), style=wx.BRUSHSTYLE_SOLID))
        gc.DrawRectangle(x * px * gc_scale, pos * py * gc_scale, w * px * gc_scale, si * py * gc_scale)

    def draw_polygon_gc(self, gc: wx.GraphicsContext, points, fill=LColor(1, 1, 1)):
        gx, gy = gc.GetSize()
        gc_scale = min(gx / self.view_box[0], gy / self.view_box[1])

        gc.SetPen(wx.BLACK_PEN)
        gc.SetBrush(wx.Brush(fill.get_wx(), style=wx.BRUSHSTYLE_SOLID))

        points2d = [wx.Point2D(gc_scale * x, gc_scale * y) for x, y in points]
        points2d.append(wx.Point2D(gc_scale * points[0][0], gc_scale * points[0][1]))

        gc.DrawLines(points2d)

    def draw_centered_text(self, gc: wx.GraphicsContext, txt: str, x, y, baseline=False):
        w, h = gc.GetFullTextExtent(txt)[:2]
        if baseline:
            gc.DrawText(txt, int(x - w / 2.0), int(y))
        else:
            gc.DrawText(txt, int(x - w / 2.0), int(y - h / 2.0))

    def render_to_gc(self, gc: wx.GraphicsContext):
        px, py = self.view_box * 0.01
        bg = self.block_generator
        gx, gy = gc.GetSize()
        gc_scale = min(gx / self.view_box[0], gy / self.view_box[1])

        vscale = bg.dtotal / 85.0  # 90% of image height to be used for layers

        blocks = bg.get_blocks()

        gc.SetFont(
            wx.Font(
                int(self.fontsize * gc_scale * 0.75), wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL
            ),
            wx.BLACK,
        )

        if self.use3d:
            tc = 45
        else:
            tc = 50

        pos = 90.0
        self.add_rect_gc(gc, pos, 100, LColor(0, 0, 0))
        bc = LColor(0, 0, 0)
        font_center = 0.
        font_offset = gc_scale*self.fontsize*1.2
        if self.show_names:
            font_center += font_offset/2.
        if self.show_slds:
            font_center -= font_offset/2.
        for bi in blocks:
            if isinstance(bi, list):
                block_length = sum([bij.scale / vscale for bij in bi])
                self.add_rect_gc(
                    gc, pos - block_length, block_length, LColor(0.75, 0.75, 0.75), width=10 + self.fontsize
                )
                for bij in bi:
                    dij = bij.scale / vscale
                    pos -= dij
                    if bij.thickness < 0:
                        if self.use3d:
                            # add surface polygon
                            shifted = self.to_vp((85, pos + dij))
                            points = [
                                ((5 + self.fontsize) * px, (pos + dij) * py),
                                (85 * px, (pos + dij) * py),
                                shifted,
                                (shifted[0] - (85 - 5 - self.fontsize) * px, shifted[1]),
                            ]
                            self.draw_polygon_gc(gc, points, bc)
                        self.draw_centered_text(gc, "...", gc_scale * tc * px, gc_scale * (pos + dij / 3.0) * py)
                        gc.DrawText(
                            f"(x%i)" % (-bij.thickness), gc_scale * 5, gc_scale * (pos + dij) * py, angle=pi / 2
                        )
                    else:
                        bc = bij.color
                        self.add_rect_gc(gc, pos, dij, bc, in_stack=True)
                        if self.show_names:
                            self.draw_centered_text(gc, bij.name, gc_scale * tc * px,
                                                    gc_scale * (pos + dij / 2.0) * py + font_center - font_offset)
                        self.draw_centered_text(gc, self.tlabel(bij.thickness),
                                                gc_scale * tc * px, gc_scale * (pos + dij / 2.0) * py + font_center
                        )
                        if self.show_slds:
                            self.draw_centered_text(gc, bij.sld, gc_scale*tc*px,
                                                    gc_scale*(pos+dij/2.0)*py + font_center + font_offset)
            else:
                di = bi.scale / vscale
                pos -= di
                bc = bi.color
                self.add_rect_gc(gc, pos, di, bc)
                if self.show_names:
                    self.draw_centered_text(gc, bi.name, gc_scale*tc*px,
                                        gc_scale*(pos+di/2.0)*py + font_center - font_offset)
                self.draw_centered_text(gc, self.tlabel(bi.thickness),
                                        gc_scale * tc * px, gc_scale * (pos + di / 2.0) * py + font_center
                )
                if self.show_slds:
                    self.draw_centered_text(gc, bi.sld, gc_scale*tc*px,
                                            gc_scale*(pos+di/2.0)*py + font_center + font_offset)
        if self.use3d:
            # add surface polygon
            shifted = self.to_vp((85, pos))
            points = [(5 * px, pos * py), (85 * px, pos * py), shifted, (shifted[0] - 80 * px, shifted[1])]
            self.draw_polygon_gc(gc, points, bc)


class SVGPanel(wx.Panel):
    # svg_img: SVGimage = None
    svg_img: SVGenerator = None
    last_scale = 1.0

    def __init__(self, parent):
        super().__init__(parent=parent, style=wx.FULL_REPAINT_ON_RESIZE)
        self.Bind(wx.EVT_PAINT, self.OnPaint)

    def OnPaint(self, event: wx.PaintEvent):
        if self.svg_img is not None:
            img = self.svg_img

            dc = wx.PaintDC(self)
            dc.SetBackground(wx.Brush("white"))
            dc.Clear()

            ctx = wx.GraphicsContext.Create(dc)
            img.render_to_gc(ctx)
            # img.RenderToGC(ctx, scale)
        else:
            event.Skip()


class Plugin(framework.Template):
    svg: str = None

    def __init__(self, parent):
        framework.Template.__init__(self, parent)
        self.parent = parent

        # Create the SLD plot
        LG_panel = self.NewPlotFolder("Layer Graphics")
        SA_sizer = wx.BoxSizer(wx.VERTICAL)
        LG_panel.SetSizer(SA_sizer)

        self.img = SVGPanel(LG_panel)
        SA_sizer.Add(self.img, 1, wx.EXPAND | wx.GROW | wx.ALL)

        bot_sizer = wx.BoxSizer(wx.HORIZONTAL)
        SA_sizer.Add(bot_sizer, 0, wx.FIXED_MINSIZE)
        left_sizer = wx.BoxSizer(wx.VERTICAL)
        bot_sizer.Add(left_sizer, 1, wx.EXPAND | wx.GROW | wx.FIXED_MINSIZE)
        bot_sizer.AddSpacer(4)
        mid_sizer = wx.BoxSizer(wx.VERTICAL)
        bot_sizer.Add(mid_sizer, 1, wx.EXPAND | wx.GROW | wx.FIXED_MINSIZE)
        bot_sizer.AddSpacer(4)
        right_sizer = wx.BoxSizer(wx.VERTICAL)
        bot_sizer.Add(right_sizer, 1, wx.EXPAND | wx.GROW | wx.FIXED_MINSIZE)

        self.rescale = wx.CheckBox(LG_panel, label="Rescale thickness display")
        self.rescale.SetValue(True)
        left_sizer.Add(self.rescale, 0, wx.FIXED_MINSIZE)

        self.show_all = wx.RadioButton(LG_panel, label="All layers")
        left_sizer.Add(self.show_all, 0, wx.FIXED_MINSIZE)

        show_topbot = wx.RadioButton(LG_panel, label="Top/Bottom layers")
        show_topbot.SetValue(True)
        left_sizer.Add(show_topbot, 0, wx.FIXED_MINSIZE)

        self.show_one = wx.RadioButton(LG_panel, label="Single Repetition")
        left_sizer.Add(self.show_one, 0, wx.FIXED_MINSIZE)

        self.show_names = wx.CheckBox(LG_panel, label="Show Names")
        self.show_names.SetValue(False)
        left_sizer.Add(self.show_names, 0, wx.FIXED_MINSIZE)

        self.show_slds = wx.CheckBox(LG_panel, label="Show SLDs")
        self.show_slds.SetValue(False)
        mid_sizer.Add(self.show_slds, 0, wx.FIXED_MINSIZE)

        self.use3d = wx.CheckBox(LG_panel, label="Pseudo 3d")
        self.use3d.SetValue(True)
        mid_sizer.Add(self.use3d, 0, wx.FIXED_MINSIZE)
        self.unit_nm = wx.CheckBox(LG_panel, label="nm-unit")
        self.unit_nm.SetValue(True)
        mid_sizer.Add(self.unit_nm, 0, wx.FIXED_MINSIZE)
        self.unit_precision = wx.SpinCtrl(LG_panel, value="1", min=0, max=5)
        mid_sizer.Add(self.unit_precision, 0, wx.FIXED_MINSIZE)

        self.fontsize = wx.SpinCtrl(LG_panel, value="10", min=1, max=20)
        right_sizer.Add(self.fontsize, 0, wx.FIXED_MINSIZE)
        export_button = wx.Button(LG_panel, label="Save to SVG...")
        right_sizer.Add(export_button)
        copy_button = wx.Button(LG_panel, label="Copy to Clipboard")
        right_sizer.Add(copy_button)

        LG_panel.Layout()
        try:
            self.OnSimulate(None)
        except Exception:
            pass

        self.rescale.Bind(wx.EVT_CHECKBOX, self.OnSimulate)
        self.show_names.Bind(wx.EVT_CHECKBOX, self.OnSimulate)
        self.show_slds.Bind(wx.EVT_CHECKBOX, self.OnSimulate)
        self.show_all.Bind(wx.EVT_RADIOBUTTON, self.OnSimulate)
        show_topbot.Bind(wx.EVT_RADIOBUTTON, self.OnSimulate)
        self.show_one.Bind(wx.EVT_RADIOBUTTON, self.OnSimulate)
        self.use3d.Bind(wx.EVT_CHECKBOX, self.OnSimulate)
        self.unit_nm.Bind(wx.EVT_CHECKBOX, self.OnSimulate)
        self.unit_precision.Bind(wx.EVT_SPINCTRL, self.OnSimulate)
        self.fontsize.Bind(wx.EVT_SPINCTRL, self.OnSimulate)
        export_button.Bind(wx.EVT_BUTTON, self.ExportSVG)
        copy_button.Bind(wx.EVT_BUTTON, self.CopyImage)

    def OnSimulate(self, event):
        # Calculate and update the sld plot
        model = self.GetModel()
        if model.script_module is None:
            return
        # create a mapping of layer object ids to layer name
        layer_names = dict([(id(item), ii)
                       for ii,item in model.script_module.__dict__.items()
                       if type(item).__name__=='Layer'])
        gen = SVGenerator(
            model.script_module.sample,
            layer_names = layer_names,
            show_names = self.show_names.GetValue(),
            show_slds = self.show_slds.GetValue(),
            rescale=self.rescale.GetValue(),
            show_all=self.show_all.GetValue(),
            show_one=self.show_one.GetValue(),
            use3d=self.use3d.GetValue(),
            unit_nm=self.unit_nm.GetValue(),
            unit_precision=int(self.unit_precision.GetValue()),
            fontsize=self.fontsize.GetValue(),
            is_xray=model.script_module.inst.probe=='x-ray',
        )
        self.svg = gen.svg.tostring()
        # self.img.svg_img = SVGimage.CreateFromBytes(self.svg.encode('utf-8'))
        self.img.svg_img = gen
        self.img.Refresh()

    def ExportSVG(self, event):
        dlg = wx.FileDialog(
            self.img,
            message="Save layer sketch image",
            defaultFile="",
            wildcard="SVG Image|*.svg",
            style=wx.FD_SAVE | wx.FD_CHANGE_DIR | wx.FD_OVERWRITE_PROMPT,
        )
        if dlg.ShowModal() == wx.ID_OK:
            path = dlg.GetPath()
            with open(path, "w", encoding="utf-8") as fh:
                fh.write(self.svg)
        dlg.Destroy()

    def CopyImage(self, event):
        bmp = wx.Bitmap(wx.Size(300, 600), depth=32)

        memdc = wx.MemoryDC(bmp)
        memdc.SetBackground(wx.Brush(wx.Colour(255, 255, 255, 0)))
        memdc.Clear()

        ctx = wx.GraphicsContext.Create(memdc)

        self.img.svg_img.render_to_gc(ctx)
        memdc.SelectObject(wx.NullBitmap)

        bmp_obj = wx.BitmapDataObject()
        bmp_obj.SetBitmap(bmp)

        if not wx.TheClipboard.IsOpened():
            open_success = wx.TheClipboard.Open()
            if open_success:
                wx.TheClipboard.SetData(bmp_obj)
                wx.TheClipboard.Close()
                wx.TheClipboard.Flush()



"""
Use Levenberg-Marquardt as minimizer to estimate errors and for fast decent to next local minimum.
"""

import _thread
import pickle
import sys

from dataclasses import dataclass

from numpy import array, ndarray, sign, sqrt
from scipy.optimize import leastsq

from .core.config import BaseConfig
from .core.custom_logging import iprint
from .exceptions import ErrorBarError, OptimizerInterrupted
from .model import Model
from .solver_basis import GenxOptimizer, GenxOptimizerCallback, SolverParameterInfo, SolverResultInfo, SolverUpdateInfo


class LMDefaultCallbacks(GenxOptimizerCallback):

    def text_output(self, text):
        iprint(text)
        sys.stdout.flush()

    def plot_output(self, update_data):
        pass

    def parameter_output(self, param_info):
        pass

    def fitting_ended(self, result_data):
        pass

    def autosave(self):
        pass


@dataclass
class LMConfig(BaseConfig):
    section = "solver"

    groups = {}


class LMOptimizer(GenxOptimizer):
    """
    Optimizer based on Levenberg-Marquardt algorithm.
    """

    opt: LMConfig
    model: Model
    fom_log: ndarray
    start_guess: ndarray
    covar: ndarray

    _callbacks: GenxOptimizerCallback = LMDefaultCallbacks()

    n_fom_evals = 0

    def is_running(self):
        return False

    def __init__(self):
        GenxOptimizer.__init__(self)
        self.model = Model()
        self.fom_log = array([[0, 0]])[0:0]
        self.covar = array([])
        self.start_guess = array([])

    def pickle_string(self, clear_evals: bool = False):
        return pickle.dumps(self)

    def pickle_load(self, pickled_string: bytes):
        obj = pickle.loads(pickled_string, encoding="latin1", errors="ignore")
        # TODO: set own options from object

    def get_start_guess(self):
        return self.start_guess

    def get_model(self) -> Model:
        return self.model

    def get_fom_log(self):
        return self.fom_log

    def connect_model(self, model_obj: Model):
        """
        Connects the model [model] to this object. Retrives the function
        that sets the variables  and stores a reference to the model.
        """
        # Retrieve parameters from the model
        (param_funcs, start_guess, par_min, par_max) = model_obj.get_fit_pars(use_bounds=False)

        # Control parameter setup
        self.par_funcs = param_funcs
        self.model = model_obj
        self.n_dim = len(param_funcs)
        self.start_guess = start_guess

    def calc_sim(self, vec):
        """calc_sim(self, vec) --> None
        Function that will evaluate the the data points for
        parameters in vec.
        """
        model_obj = self.model
        # Set the parameter values
        list(map(lambda func, value: func(value), self.par_funcs, vec))

        self.model.evaluate_sim_func()
        return self.model.fom

    def calc_fom(self, vec):
        """
        Function to calcuate the figure of merit for parameter vector
        vec.
        """
        if self._stop_fit:
            raise OptimizerInterrupted("interrupted")
        # Set the parameter values
        list(map(lambda func, value: func(value), self.par_funcs, vec))
        fom = self.model.evaluate_fit_func(get_elements=True)  # fom is squared in leastsq
        chi = sign(fom) * sqrt(abs(fom))
        self.n_fom_evals += 1
        return chi

    def calc_error_bar(self, index: int) -> (float, float):
        if self.covar is None:
            raise ErrorBarError(
                "Could not get covariance matrix from fit, maybe the parameters are coupled/have no influence?"
            )
        err = sqrt(self.covar[index, index])
        return -err, err

    def project_evals(self, index: int):
        # -> (ArrayLike, ArrayLike)
        pass

    def start_fit(self, model: Model):
        self.n_fom_evals = 0
        self.connect_model(model)
        self._stop_fit = False
        _thread.start_new_thread(self.optimize, ())

    def optimize(self):
        try:
            res = leastsq(self.calc_fom, self.start_guess, full_output=True)
        except OptimizerInterrupted:
            self._callbacks.fitting_ended(self.get_result_info(interrupted=True))
            return
        self.best_vec = res[0]
        if res[1] is None:
            self.covar = None
        else:
            Chi2Res = self.calc_fom(self.best_vec) ** 2
            s_sq = Chi2Res.sum() / (len(Chi2Res) - len(res[0]))  # variance of the residuals
            self.covar = res[1] * s_sq

        self.plot_output()
        self._callbacks.fitting_ended(self.get_result_info())

    def stop_fit(self):
        self._stop_fit = True

    def resume_fit(self, model: Model):
        pass

    def is_fitted(self):
        return self.n_fom_evals > 0

    def is_configured(self) -> bool:
        pass

    def set_callbacks(self, callbacks: GenxOptimizerCallback):
        self._callbacks = callbacks

    def get_callbacks(self) -> "GenxOptimizerCallback":
        return self._callbacks

    def plot_output(self):
        self.calc_sim(self.best_vec)
        data = SolverUpdateInfo(
            fom_value=self.model.fom,
            fom_name=self.model.fom_func.__name__,
            fom_log=self.get_fom_log(),
            new_best=True,
            data=self.model.data,
        )
        self._callbacks.plot_output(data)

    def get_result_info(self, interrupted=False):
        result = SolverResultInfo(
            start_guess=self.start_guess.copy(),
            error_message="",
            values=self.best_vec.copy(),
            new_best=not interrupted,
            population=[],
            max_val=[],
            min_val=[],
            fitting=False,
        )
        return result



"""
Configuration and config file handling.
"""

import dataclasses
import io
import os

from abc import ABC, abstractmethod
from configparser import ConfigParser
from functools import lru_cache
from logging import debug
from typing import Dict, List, Type, get_type_hints

from ..exceptions import GenxIOError, GenxOptionError
from .custom_logging import iprint


# ==============================================================================
class Config:
    """
    GenX configuration handler.
    """

    def __init__(self):
        self.default_config = ConfigParser()
        self.model_config = ConfigParser()

    def load_default(self, filename, reset=False):
        """
        Loads the default config from file filename. Raises a IOError if the
        can not be found.
        """
        if not os.path.exists(filename):
            raise GenxIOError(f"Could not load default config file {filename}:\n File does not exist", filename)
        try:
            with open(filename, "r") as f:
                f.readline()
        except Exception as e:
            iprint(e)
            raise GenxIOError(f"Could not load default config file:\n {e}", filename)

        # the read method ignores any errors, thus checks have to be done in this method beforehand
        self.default_config.read(filename)

        if reset:
            self.model_config = ConfigParser()

    def write_default(self, filename):
        """
        Writes the current default configuration to filename
        """
        try:
            with open(filename, "w") as cfile:
                self.default_config.write(cfile)
        except Exception as e:
            iprint(e)
            raise GenxIOError(f"Could not write default config file:\n {e}", filename)

    def load_string(self, string: str):
        """
        Loads a config from a string input.  Raises an IOError if the string can not be
        read.
        """
        try:
            self.model_config.read_string(string)
        except Exception as e:
            raise GenxIOError(f"Could not load model config file:\n {e}")

    def _getf(self, method: str, section: str, option: str, fallback=None):
        """
        For the getter-method name of config objects try to locate the section and option first in
        model_config if that fails try to locate it in default_config. If both
        fails raise a GenxOptionError.
        """
        if self.model_config.has_option(section, option):
            func = getattr(self.model_config, method)
        elif self.default_config.has_option(section, option):
            func = getattr(self.default_config, method)
        else:
            raise GenxOptionError(section, option)
        try:
            return func(section, option, fallback=fallback)
        except ValueError:
            if fallback is None:
                raise GenxOptionError(section, option)
            else:
                debug(f"Return fallback due to ValueError in _getf, section={section}/option={option}")
                return fallback

    def getlist(self, section: str, option: str, fallback=None):
        str_value = self.get(section, option, fallback=fallback)
        return [s.strip() for s in str_value.split(";")]

    @lru_cache(maxsize=10)
    def __getattr__(self, item: str):
        """
        Handle all get functions from ConfigParser to be redirected through _getf,
        everything else is treated normally.
        The lambda functions are cached so they only get generated once.
        """
        if item.startswith("get") and hasattr(self.model_config, item):
            return lambda section, option, fallback=None: self._getf(item, section, option, fallback=fallback)
        else:
            raise AttributeError(f"'Config' object has no attribute '{item}'")

    def model_set(self, section, option, value):
        """
        Set a value in section, option for the model configuration.
        """
        if not self.model_config.has_section(section):
            self.model_config.add_section(section)
        self.model_config.set(section, option, str(value))

    def default_set(self, section, option, value):
        """
        Set a value in section, option for the model configuration.
        """
        if not self.default_config.has_section(section):
            self.default_config.add_section(section)
        self.default_config.set(section, option, str(value))

    def set(self, section, option, value):
        """
        Set a value in section, option for the model configuration.
        Just a duplication of model_set.
        """
        self.model_set(section, option, value)

    def model_dump(self):
        """
        dumps the model configuration to a string.
        """
        buffer = io.StringIO()
        self.model_config.write(buffer)
        string = buffer.getvalue()
        buffer.close()
        return string


# create the global GenX config object used by the program, it does not depend on anything else
config = Config()


class BaseConfig(ABC):
    """
    Base class for handling configuration that shall be stored by a GenX Config object.

    Handles general parameter read/write so that derived classes only have to
    inherit from it and use dataclass attributes.
    """

    section: str

    @property
    @abstractmethod
    def section(self):
        """Subclass needs to define the section attribute."""

    def asdict(self):
        # noinspection PyDataclass
        return dataclasses.asdict(self)

    def load_config(self):
        field: dataclasses.Field
        # noinspection PyDataclass
        for field in dataclasses.fields(self):
            if field.type is float:
                getf = config.getfloat
            elif field.type is int:
                getf = config.getint
            elif field.type is bool:
                getf = config.getboolean
            elif field.type is list:
                getf = config.getlist
            else:
                getf = config.get

            if field.default is dataclasses.MISSING:
                fallback = None
            else:
                fallback = field.default
            option = field.name.replace("_", " ")
            try:
                value = getf(self.section, option, fallback=fallback)
            except GenxOptionError as e:
                debug(f"Could not read option {self.section}/{option} of type {field.type.__name__}:\n    {e}")
            else:
                setattr(self, field.name, value)

    def save_config(self, default=False):
        data = self.asdict()
        if default:
            setter = config.default_set
        else:
            setter = config.model_set
        for key, value in data.items():
            option = key.replace("_", " ")
            if type(value) is list:
                value = ";".join(value)
            setter(self.section, option, value)

    def copy(self):
        # noinspection PyDataclass
        return dataclasses.replace(self)

    @property
    def groups(self) -> Dict[str, List[str]]:
        """
        Can be replaced by dictionary in sub-classes to define how to group parameters in display and
        configuration dialog entries.
        If a list entry contains a list of strings it is used to group the values horizontally.
        A list of 4 items with the second beeing a boolean is interpreted as a choice between two parameters:
            [{Choice Group Label}, bool-first is active, first value, second value]
        """
        res = list(self.asdict().keys())
        return {"": res}

    def get_fields(self, fltr=None) -> List[dataclasses.Field]:
        """
        Return the fields for the dataclass.
        If fltr is supplied the results will be filtered by field names given in this list.
        """
        # noinspection PyDataclass
        fields = dataclasses.fields(self)  # child classes will be dataclass objects
        if fltr is None:
            return list(fields)
        else:
            out = []
            for fi in fields:
                if fi.name in fltr:
                    out.append(fi)
            return out

    @staticmethod
    def GParam(default, pmin=None, pmax=None, label=None, descriptoin=None):
        # allow to add some metadata to the parameter to use in dialog generation
        gmeta = {}
        if pmin is not None:
            gmeta["pmin"] = pmin
        if pmax is not None:
            gmeta["pmax"] = pmax
        if label is not None:
            gmeta["label"] = label
        if descriptoin is not None:
            gmeta["descriptoin"] = descriptoin
        return dataclasses.field(default=default, metadata={"genx": gmeta})

    @staticmethod
    def GChoice(default, selection, label=None, descriptoin=None):
        """
        Allow to add some metadata to the parameter to use in dialog generation,
        selection is a list of alloweed string values for this parameter.
        """
        gmeta = {"selection": selection}
        if label is not None:
            gmeta["label"] = label
        if descriptoin is not None:
            gmeta["descriptoin"] = descriptoin
        return dataclasses.field(default=default, metadata={"genx": gmeta})

    def __or__(self, other):
        return MergedConfig(self, other)


class MergedConfig(BaseConfig):
    """
    A config-like object that combines two or more config objects.

    Attribut access is passed to the child objects.
    """

    section = None
    _children: List[BaseConfig]

    def __init__(self, *children):
        self._children = list(children)

    def __eq__(self, other):
        if not hasattr(other, "_children") or len(other._children) != len(self._children):
            return False
        out = True
        for i, child in enumerate(self._children):
            out &= child.__eq__(other._children[i])
        return out

    def __or__(self, other):
        if hasattr(other, "_children"):
            all_children = self._children + other._children
        else:
            all_children = self._children + [other]
        return MergedConfig(*all_children)

    def get_fields(self, fltr=None) -> List[dataclasses.Field]:
        fout = []
        for c in self._children:
            fout += c.get_fields(fltr)
        return fout

    def copy(self):
        clist = [c.copy() for c in self._children]
        return MergedConfig(*clist)

    @property
    def groups(self) -> Dict[str, List[str]]:
        dout = {}
        for c in self._children:
            dout.update(c.groups)
        return dout

    def __getattr__(self, name):
        # search through the child configuration fields if name is in there and return value
        for c in self._children:
            fnames = [fi.name for fi in c.get_fields()]
            if name in fnames:
                return getattr(c, name)
        raise AttributeError("{item} not found in any child configuration")

    def __setattr__(self, name, value):
        if name.startswith("_"):
            BaseConfig.__setattr__(self, name, value)
        # search through the child configuration fields if name is in there, otherwise set own attribute
        for c in self._children:
            fnames = [fi.name for fi in c.get_fields()]
            if name in fnames:
                return setattr(c, name, value)
        BaseConfig.__setattr__(self, name, value)


class Configurable:
    """
    A mixin class for classes that store parameters in a savable configuration.

    Defines methods for reading and writing configurations. Subclasses can
    define the config_updated method to be called after reading a configuration.

    If no config_class is passed to the constructor explicitly, any existing type hint
    to a BaseConfig derived object will be used to configure the opt attribute.
    """

    def __init__(self, config_class: Type[BaseConfig] = None):
        if config_class is None:
            hints = get_type_hints(self.__class__)
            config_class = hints.get("opt", None)

        if not issubclass(config_class, BaseConfig):
            raise ValueError(
                "Configurable needs either an explicit config_class or 'opt' type hint derived from BaseConfig"
            )

        self.opt = config_class()

    def ReadConfig(self):
        """Reads the variables stored in the config file."""
        self.opt.load_config()
        self.UpdateConfigValues()

    def WriteConfig(self, default=False):
        """Writes the varaibles to be stored to the config"""
        self.opt.save_config(default=default)

    def UpdateConfigValues(self):
        """
        Sub-classes can overwrite this method to preform some action after
        a configuration has been loaded.
        """



"""
Single-instance socket helpers for the GenX GUI.
"""

from __future__ import annotations

import json
import socket
import threading
from typing import Callable, List

import wx


class SingleInstanceServer(threading.Thread):
    """
    Background server thread listening for single-instance client messages.
    """

    HANDSHAKE_TOKEN = "GENX_SINGLE_INSTANCE_HELLO"
    _RESPONSE_OK = "OK"
    _RESPONSE_ACK = "ACK"
    _RESPONSE_ERR = "ERR"

    def __init__(
        self,
        on_message: Callable[[List[str]], None] | None = None,
        host: str = "127.0.0.1",
        port: int = 37983,
        timeout: float = 0.5,
    ) -> None:
        super().__init__(name="GenXSingleInstanceServer", daemon=True)
        self.on_message = on_message
        self.host = host
        self.port = port
        self.timeout = timeout
        self._stop_event = threading.Event()
        self._socket: socket.socket | None = None

    def stop(self) -> None:
        self._stop_event.set()
        if self._socket is not None:
            try:
                self._socket.close()
            except OSError:
                pass
            self._socket = None
        self.join(1.0)

    def run(self) -> None:
        server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._socket = server
        server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server.bind((self.host, self.port))
        server.listen(1)
        server.settimeout(self.timeout)

        while not self._stop_event.is_set():
            try:
                conn, _ = server.accept()
            except socket.timeout:
                continue
            except OSError:
                break
            with conn:
                self._handle_client(conn)

        try:
            server.close()
        except OSError:
            pass

    def _handle_client(self, conn: socket.socket) -> None:
        conn.settimeout(5.0)
        handshake = self._recv_line(conn)
        if handshake != self.HANDSHAKE_TOKEN:
            self._send_line(conn, self._RESPONSE_ERR)
            return
        self._send_line(conn, self._RESPONSE_OK)

        payload = self._recv_line(conn)
        try:
            data = json.loads(payload)
        except json.JSONDecodeError:
            self._send_line(conn, self._RESPONSE_ERR)
            return

        if not isinstance(data, dict):
            self._send_line(conn, self._RESPONSE_ERR)
            return

        messages = data.get("messages")
        if not isinstance(messages, list) or not all(isinstance(item, str) for item in messages):
            self._send_line(conn, self._RESPONSE_ERR)
            return

        if self.on_message is not None:
            wx.CallAfter(self.on_message, messages)

        self._send_line(conn, self._RESPONSE_ACK)

    @staticmethod
    def _recv_line(conn: socket.socket) -> str:
        data = bytearray()
        while True:
            chunk = conn.recv(1)
            if not chunk:
                break
            if chunk == b"\n":
                break
            data.extend(chunk)
        return data.decode("utf-8")

    @staticmethod
    def _send_line(conn: socket.socket, message: str) -> None:
        conn.sendall(message.encode("utf-8") + b"\n")


class SingleInstanceClient:
    """
    Client for checking a running single-instance server and sending messages.
    """

    def __init__(self, host: str = "127.0.0.1", port: int = 37983, timeout: float = 1.0) -> None:
        self.host = host
        self.port = port
        self.timeout = timeout

    def is_server_running(self) -> bool:
        try:
            with socket.create_connection((self.host, self.port), timeout=0.05):
                return True
        except OSError:
            return False

    def send_message(self, messages: List[str]) -> bool:
        if not isinstance(messages, list) or not all(isinstance(item, str) for item in messages):
            raise TypeError("messages must be a list[str]")

        try:
            with socket.create_connection((self.host, self.port), timeout=self.timeout) as conn:
                conn.settimeout(self.timeout)

                self._send_line(conn, SingleInstanceServer.HANDSHAKE_TOKEN)
                response = self._recv_line(conn)
                if response != SingleInstanceServer._RESPONSE_OK:
                    return False

                payload = json.dumps({"messages": messages}, ensure_ascii=True)
                self._send_line(conn, payload)
                ack = self._recv_line(conn)
                return ack == SingleInstanceServer._RESPONSE_ACK
        except OSError:
            return False


    @staticmethod
    def _recv_line(conn: socket.socket) -> str:
        data = bytearray()
        while True:
            chunk = conn.recv(1)
            if not chunk:
                break
            if chunk == b"\n":
                break
            data.extend(chunk)
        return data.decode("utf-8")

    @staticmethod
    def _send_line(conn: socket.socket, message: str) -> None:
        conn.sendall(message.encode("utf-8") + b"\n")



"""
==========================================
:mod:`orso` ORSO standard file data loader
==========================================

Loads the format following the specification of the Open Reflectometry Standards Organization (ORSO).
See https://www.reflectometry.org/working_groups/file_formats/ for more details
"""

import re

from typing import List

import h5py
import numpy as np

from orsopy.fileio import OrsoDataset, load_nexus, load_orso

from ..data_loader_framework import Template
from ..utils import ShowWarningDialog


class Plugin(Template):
    wildcard = "*.ort;*.orb;*.nxs;*.h5;*.hdf"
    _cached_file = ""
    _cached_data: List[OrsoDataset] = None

    def __init__(self, parent):
        Template.__init__(self, parent)
        self.x_col = 0
        self.y_col = 1
        self.e_col = 3
        self.xe_col = 2
        self.comment = "#"
        self.skip_rows = 0
        self.delimiter = None

    def CanOpen(self, file_path):
        if not Template.CanOpen(self, file_path):
            return False
        if (
            file_path.endswith(".orb")
            or file_path.endswith(".nxs")
            or file_path.endswith(".h5")
            or file_path.endswith(".hdf")
        ):
            return h5py.is_hdf5(file_path)
        else:
            with open(file_path, "r", encoding="utf-8") as fh:
                l1 = fh.readline()
            return l1.startswith("# # ORSO")

    def LoadCached(self, file_path):
        if self._cached_file == file_path:
            return self._cached_data
        else:
            if (
                file_path.endswith(".orb")
                or file_path.endswith(".nxs")
                or file_path.endswith(".h5")
                or file_path.endswith(".hdf")
            ):
                orso_datasets = load_nexus(file_path)
            else:
                orso_datasets = load_orso(file_path)
            self._cached_file = file_path
            self._cached_data = orso_datasets
            return orso_datasets

    def CountDatasets(self, file_path):
        try:
            orso_datasets = self.LoadCached(file_path)
        except Exception as e:
            return 1
        else:
            return len(orso_datasets)

    def LoadData(self, dataset, filename, data_id=0):
        """
        Loads the data from filename into the data_item_number.
        """
        try:
            orso_datasets = self.LoadCached(filename)
        except Exception as e:
            ShowWarningDialog(
                self.parent,
                "Could not load the file: "
                + filename
                + " \nPlease check the format.\n\n ORSO reader gave the following error:\n"
                + str(e),
            )
            return
        else:
            orso_dataset: OrsoDataset = orso_datasets[data_id]
            data: np.ndarray = orso_dataset.data.T

            cols = orso_dataset.info.columns

            start_usercols = 3
            dataset.x_raw = np.asarray(data[0])
            if orso_dataset.info.columns[0].unit == "1/nm":
                dataset.x_raw /= 10.0
            dataset.y_raw = np.asarray(data[1])
            dataset.error_raw = np.asarray(data[2])
            if data.shape[0] > 3 and getattr(cols[3], "error_of", None) == cols[0].name:
                start_usercols = 4
                dataset.set_extra_data("res", np.asarray(data[3]) * cols[3].to_sigma, "res")
            for i, col in enumerate(data[start_usercols:]):
                colname = cols[i + start_usercols].name
                if not colname.isidentifier() or colname in ["lambda", "x", "y", "error", "res"]:
                    colname = "col_" + re.sub("[^0-9a-zA-Z_]", "", colname)
                dataset.set_extra_data(colname, np.asarray(col), colname)
            # Name the dataset accordign to file name
            if orso_dataset.info.data_set:
                dataset.name = str(orso_dataset.info.data_set)
            # Run the commands on the data - this also sets the x,y, error memebers
            # of that data item.
            dataset.run_command()

            # insert metadata into ORSO compatible fields
            dataset.meta = orso_dataset.info.to_dict()
            if "data_set" in dataset.meta:
                del dataset.meta["data_set"]
